local ____lualib = require("lualib_bundle")
local __TS__ArrayMap = ____lualib.__TS__ArrayMap
local ____exports = {}
local ____lib = require("lib.index")
local getKeySet = ____lib.getKeySet
local isEmpty = ____lib.isEmpty
local ____geometry = require("lib.geometry.index")
local Pos = ____geometry.Pos
____exports.FirstEntityOriginalPositionTag = "bp100:FirstEntityOriginalPosition"
local function adjustEntitiesToMatchPositionOffset(stack, entities, positionOffset, firstEntityOriginalPosition)
    local firstEntityPosition = entities[1].position
    local expectedPosition = Pos.plus(firstEntityOriginalPosition, positionOffset)
    local shouldAdjustPosition = not Pos.equals(firstEntityPosition, expectedPosition)
    if not shouldAdjustPosition then
        return false
    end
    local ____Pos_minus_result_0 = Pos.minus(expectedPosition, firstEntityPosition)
    local x = ____Pos_minus_result_0.x
    local y = ____Pos_minus_result_0.y
    for i = 1, #entities do
        local pos = entities[i].position
        pos.x = pos.x + x
        pos.y = pos.y + y
    end
    local tiles = stack.get_blueprint_tiles()
    if tiles then
        for i = 1, #tiles do
            local pos = tiles[i].position
            pos.x = pos.x + x
            pos.y = pos.y + y
        end
        stack.set_blueprint_tiles(tiles)
    end
    return true
end
local function getEffectivePositionOffset(entities, firstEntityOriginalPosition)
    return Pos.minus(entities[1].position, firstEntityOriginalPosition)
end
local function filterEntities(entities, unitNumbers, bpMapping, additionalWhitelist, blacklist)
    local anyDeleted
    local firstNotDeletedLuaIndex
    local un
    for i = 1, #entities do
        local entity = entities[i]
        local name = entity.name
        local ____temp_2 = blacklist[name] ~= nil
        if not ____temp_2 then
            local ____temp_1 = unitNumbers and not (additionalWhitelist[name] ~= nil)
            if ____temp_1 then
                un = bpMapping[i].unit_number
                ____temp_1 = not (un and unitNumbers[un] ~= nil)
            end
            ____temp_2 = ____temp_1
        end
        if ____temp_2 then
            entities[i] = nil
            anyDeleted = true
        else
            if firstNotDeletedLuaIndex == nil then
                firstNotDeletedLuaIndex = i
            end
        end
    end
    return anyDeleted, firstNotDeletedLuaIndex
end
local function getInfinityEntityNames()
    local infinityChestNames = getKeySet(game.get_filtered_entity_prototypes({{filter = "type", type = "infinity-container"}}))
    local infinityPipeNames = getKeySet(game.get_filtered_entity_prototypes({{filter = "type", type = "infinity-pipe"}}))
    return infinityChestNames, infinityPipeNames
end
local function replaceInfinityEntitiesWithCombinators(entities)
    local anyReplaced = false
    local chests, pipes = getInfinityEntityNames()
    for ____, entity in pairs(entities) do
        local name = entity.name
        if chests[name] ~= nil then
            local ____opt_3 = entity.infinity_settings
            local infinityFilters = ____opt_3 and ____opt_3.filters
            entity.name = "constant-combinator"
            entity.control_behavior = {filters = infinityFilters and __TS__ArrayMap(
                infinityFilters,
                function(____, f) return {index = f.index, count = f.count or 1, signal = {type = "item", name = f.name}} end
            )}
            entity.infinity_settings = nil
            anyReplaced = true
        elseif pipes[name] ~= nil then
            local settings = entity.infinity_settings
            entity.name = "constant-combinator"
            entity.control_behavior = {filters = settings and ({{
                index = 1,
                count = math.ceil((settings.percentage or 0.01) * 100),
                signal = {type = "fluid", name = settings.name}
            }})}
        end
        anyReplaced = true
    end
    return anyReplaced
end
local function contains2x2Grid(bp)
    bp.blueprint_snap_to_grid = {1, 1}
    local ____bp_blueprint_snap_to_grid_5 = bp.blueprint_snap_to_grid
    local x = ____bp_blueprint_snap_to_grid_5.x
    local y = ____bp_blueprint_snap_to_grid_5.y
    return x == 2 and y == 2
end
local function posIsOdd(pos)
    return pos.x % 2 == 1 or pos.y % 2 == 1
end
local function alignPosTo2x2(pos, isOdd)
    local rem = isOdd and 1 or 0
    return {
        x = math.floor(pos.x / 2) * 2 + rem,
        y = math.floor(pos.y / 2) * 2 + rem
    }
end
--- If forEdit is true, sets the first entity's original position tag.
function ____exports.takeSingleBlueprint(stack, params, surface, bbox, unitNumberFilter, forEdit)
    stack.set_stack("blueprint")
    stack.clear_blueprint()
    local bpMapping = stack.create_blueprint({
        surface = surface,
        force = "player",
        area = bbox,
        include_trains = true,
        include_station_names = true,
        always_include_tiles = true
    })
    if isEmpty(bpMapping) then
        return nil
    end
    local ____params_6 = params
    local snapToGrid = ____params_6.snapToGrid
    local additionalWhitelist = ____params_6.additionalWhitelist
    local blacklist = ____params_6.blacklist
    local shouldReplaceInfinity = ____params_6.replaceInfinityEntitiesWithCombinators
    local absoluteSnapping = ____params_6.absoluteSnapping
    local ____params_7 = params
    local positionOffset = ____params_7.positionOffset
    local positionRelativeToGrid = ____params_7.positionRelativeToGrid
    local entities = stack.get_blueprint_entities()
    local effectivePositionOffset
    local entitiesAdjusted = false
    local firstEntityOriginalPosition = bpMapping[1].position
    local finalFirstEntityOrigPosition = firstEntityOriginalPosition
    if snapToGrid and positionOffset then
        if contains2x2Grid(stack) then
            local isOddGrid = params.absoluteSnapping ~= nil and positionRelativeToGrid ~= nil and posIsOdd(positionRelativeToGrid)
            positionOffset = alignPosTo2x2(positionOffset, isOddGrid)
            if positionRelativeToGrid then
                positionRelativeToGrid = alignPosTo2x2(positionRelativeToGrid, isOddGrid)
            end
        end
        if adjustEntitiesToMatchPositionOffset(stack, entities, positionOffset, firstEntityOriginalPosition) then
            entitiesAdjusted = true
        end
        effectivePositionOffset = positionOffset
    else
        effectivePositionOffset = getEffectivePositionOffset(entities, firstEntityOriginalPosition)
    end
    if unitNumberFilter or blacklist then
        local anyDeleted, firstNotDeletedIndex = filterEntities(
            entities,
            unitNumberFilter,
            bpMapping,
            additionalWhitelist or ({}),
            blacklist or ({})
        )
        if anyDeleted then
            entitiesAdjusted = true
            if firstNotDeletedIndex == nil then
                stack.clear_blueprint()
                return nil
            end
            finalFirstEntityOrigPosition = bpMapping[firstNotDeletedIndex].position
        end
    end
    if shouldReplaceInfinity and replaceInfinityEntitiesWithCombinators(entities) then
        entitiesAdjusted = true
    end
    if isEmpty(entities) then
        stack.clear_blueprint()
        return nil
    end
    stack.blueprint_snap_to_grid = snapToGrid
    stack.blueprint_absolute_snapping = absoluteSnapping
    if absoluteSnapping then
        stack.blueprint_position_relative_to_grid = positionRelativeToGrid
    end
    if entitiesAdjusted then
        stack.set_blueprint_entities(entities)
    end
    stack.blueprint_icons = params.icons or stack.default_icons
    if forEdit then
        stack.set_blueprint_entity_tag(1, ____exports.FirstEntityOriginalPositionTag, finalFirstEntityOrigPosition)
    end
    return {effectivePositionOffset = effectivePositionOffset, entities = entities, bpMapping = bpMapping}
end
return ____exports
