local ____lualib = require("lualib_bundle")
local __TS__ArrayIncludes = ____lualib.__TS__ArrayIncludes
local __TS__ArrayFilter = ____lualib.__TS__ArrayFilter
local __TS__ArraySort = ____lualib.__TS__ArraySort
local __TS__ArrayMap = ____lualib.__TS__ArrayMap
local __TS__SparseArrayNew = ____lualib.__TS__SparseArrayNew
local __TS__SparseArrayPush = ____lualib.__TS__SparseArrayPush
local __TS__SparseArraySpread = ____lualib.__TS__SparseArraySpread
local ____exports = {}
local getPreviewTint, spriteToRailPieceLayers, getCurvedRailSprite, getDiagonalRailSprite, getStraightRailSprite, createRailPictures, createRailPreview, whiteTile, previewTint, utilityConstants
local util = require("util")
local ____util = require("util")
local empty_sprite = ____util.empty_sprite
local ____data_2Dutil = require("data-util")
local emptySprite16 = ____data_2Dutil.emptySprite16
local ____geometry = require("lib.geometry.index")
local BBox = ____geometry.BBox
function getPreviewTint(color)
    return color and util.mix_color(color, previewTint) or previewTint
end
function spriteToRailPieceLayers(sprite)
    return {metals = emptySprite16, backplates = emptySprite16, ties = emptySprite16, stone_path = sprite}
end
function getCurvedRailSprite(pos, size, tint)
    return spriteToRailPieceLayers({
        filename = "__bp100__/graphics/rails/curved-rail.png",
        priority = "extra-high",
        position = pos,
        size = size,
        scale = 32,
        tint = tint,
        flags = {"group=none"}
    })
end
function getDiagonalRailSprite(n, tint)
    return spriteToRailPieceLayers({
        filename = "__bp100__/graphics/rails/diagonal-rail.png",
        priority = "extra-high",
        position = {n * 2, 0},
        size = 2,
        scale = 32,
        tint = tint,
        flags = {"group=none"}
    })
end
function getStraightRailSprite(tint)
    return spriteToRailPieceLayers({
        filename = whiteTile,
        priority = "extra-high",
        size = 2,
        scale = 32,
        tint = tint,
        flags = {"group=none"}
    })
end
function createRailPictures(color)
    local tint = getPreviewTint(color)
    local straight = getStraightRailSprite(tint)
    return {
        straight_rail_horizontal = straight,
        straight_rail_vertical = straight,
        straight_rail_diagonal_left_top = getDiagonalRailSprite(0, tint),
        straight_rail_diagonal_right_top = getDiagonalRailSprite(1, tint),
        straight_rail_diagonal_right_bottom = getDiagonalRailSprite(2, tint),
        straight_rail_diagonal_left_bottom = getDiagonalRailSprite(3, tint),
        curved_rail_vertical_left_top = getCurvedRailSprite({0, 0}, {4, 8}, tint),
        curved_rail_vertical_right_top = getCurvedRailSprite({4, 0}, {4, 8}, tint),
        curved_rail_vertical_right_bottom = getCurvedRailSprite({8, 0}, {4, 8}, tint),
        curved_rail_vertical_left_bottom = getCurvedRailSprite({12, 0}, {4, 8}, tint),
        curved_rail_horizontal_left_top = getCurvedRailSprite({0, 8}, {8, 4}, tint),
        curved_rail_horizontal_right_top = getCurvedRailSprite({8, 8}, {8, 4}, tint),
        curved_rail_horizontal_right_bottom = getCurvedRailSprite({0, 12}, {8, 4}, tint),
        curved_rail_horizontal_left_bottom = getCurvedRailSprite({8, 12}, {8, 4}, tint),
        rail_endings = {sheet = emptySprite16}
    }
end
function createRailPreview(prototype, placeableBy, flags)
    local isCurved = prototype.type == "curved-rail"
    local color = prototype.friendly_map_color or prototype.map_color or utilityConstants.chart.rail_color
    if not __TS__ArrayIncludes(flags, "building-direction-8-way") then
        flags[#flags + 1] = "building-direction-8-way"
    end
    local result = {
        name = "bp100:preview-entity-" .. prototype.name,
        type = "rail-remnants",
        localised_name = {"bp100.preview-entity", {"entity-name." .. prototype.name}},
        bending_type = isCurved and "turn" or "straight",
        icons = prototype.icons,
        icon_size = prototype.icon_size,
        icon_mipmaps = prototype.icon_mipmaps,
        icon = prototype.icon,
        tile_height = prototype.tile_height,
        tile_width = prototype.tile_width,
        collision_mask = {"not-colliding-with-itself"},
        collision_box = prototype.collision_box,
        secondary_collision_box = isCurved and ({{-0.65, -2.43}, {0.65, 2.43}}) or nil,
        open_sound = prototype.open_sound,
        close_sound = prototype.close_sound,
        pictures = createRailPictures(color),
        time_before_removed = 2000000,
        remove_on_tile_placement = false,
        remove_on_entity_placement = false,
        flags = flags,
        placeable_by = placeableBy,
        final_render_layer = "ground-patch-higher2",
        subgroup = "bp100:preview-entity"
    }
    return result
end
local direction = defines.direction
local ceil = math.ceil
local max = math.max
whiteTile = "__base__/graphics/terrain/lab-tiles/lab-white.png"
previewTint = {0.5, 0.5, 0.5, 0.8}
function ____exports.createWhiteSprite(rawBBox, color)
    local bbox = BBox.normalize(rawBBox)
    local size = bbox:size()
    local scale = ceil(max(size.x, size.y))
    if scale == 0 then
        return empty_sprite()
    end
    if scale > 32 then
        local scaleX = 1
        local scaleY = 1
        if ceil(size.x) > 32 then
            scaleX = 32 / ceil(size.x)
        end
        if ceil(size.y) > 32 then
            scaleY = 32 / ceil(size.y)
        end
        bbox = bbox:scaleXY(scaleX, scaleY)
        size = bbox:size()
        scale = ceil(max(size.x, size.y))
    end
    local ____temp_0 = bbox:scale(32 / scale):size()
    local x = ____temp_0.x
    local y = ____temp_0.y
    local center = bbox:center()
    local function createRotated(dir)
        local isRotated90 = dir == direction.east or dir == direction.west
        return {
            filename = whiteTile,
            width = isRotated90 and y or x,
            height = isRotated90 and x or y,
            scale = scale,
            shift = center:rotateAboutOrigin(dir):asArray(),
            priority = "extra-high",
            flags = {"group=none"},
            tint = getPreviewTint(color)
        }
    end
    if bbox:isCenteredSquare() then
        return createRotated(direction.north)
    end
    return {
        north = createRotated(direction.north),
        east = createRotated(direction.east),
        south = createRotated(direction.south),
        west = createRotated(direction.west)
    }
end
local entityToItemBuild = {}
local itemTypes = {"item", "item-with-entity-data", "rail-planner"}
for ____, ____type in ipairs(itemTypes) do
    do
        local prototypes = data.raw[____type]
        if prototypes == nil then
            goto __continue10
        end
        for name, itemPrototype in pairs(prototypes) do
            if itemPrototype.place_result then
                entityToItemBuild[itemPrototype.place_result] = name
            end
        end
    end
    ::__continue10::
end
utilityConstants = data.raw["utility-constants"].default
local flagsToTransfer = {["placeable-off-grid"] = true, ["building-direction-8-way"] = true}
local function isBuildablePrototype(prototype)
    local flags = prototype.flags
    if not flags or not __TS__ArrayIncludes(flags, "player-creation") then
        return false
    end
    return prototype.selection_box ~= nil
end
local rollingStockTypes = {["artillery-wagon"] = true, ["cargo-wagon"] = true, ["fluid-wagon"] = true, locomotive = true}
local railPrototypes = {["straight-rail"] = true, ["curved-rail"] = true}
local previews = {}
local types = {
    "transport-belt",
    "accumulator",
    "artillery-turret",
    "beacon",
    "boiler",
    "burner-generator",
    "arithmetic-combinator",
    "decider-combinator",
    "constant-combinator",
    "container",
    "logistic-container",
    "infinity-container",
    "assembling-machine",
    "rocket-silo",
    "furnace",
    "electric-energy-interface",
    "electric-pole",
    "gate",
    "generator",
    "heat-interface",
    "heat-pipe",
    "inserter",
    "lab",
    "lamp",
    "land-mine",
    "linked-container",
    "mining-drill",
    "offshore-pump",
    "pipe",
    "infinity-pipe",
    "pipe-to-ground",
    "player-port",
    "power-switch",
    "programmable-speaker",
    "pump",
    "radar",
    "curved-rail",
    "straight-rail",
    "rail-chain-signal",
    "rail-signal",
    "reactor",
    "roboport",
    "simple-entity-with-owner",
    "simple-entity-with-force",
    "solar-panel",
    "storage-tank",
    "train-stop",
    "linked-belt",
    "loader-1x1",
    "loader",
    "splitter",
    "underground-belt",
    "turret",
    "ammo-turret",
    "electric-turret",
    "fluid-turret",
    "artillery-wagon",
    "cargo-wagon",
    "fluid-wagon",
    "locomotive",
    "wall"
}
local buildableNames = {}
for ____, ____type in ipairs(__TS__ArraySort(types)) do
    do
        local prototypes = data.raw[____type]
        if not prototypes then
            goto __continue17
        end
        for name, prototype in pairs(prototypes) do
            do
                if not isBuildablePrototype(prototype) then
                    goto __continue19
                end
                buildableNames[#buildableNames + 1] = name
                local placeableBy = prototype.placeable_by
                if not placeableBy then
                    local itemToBuild = entityToItemBuild[name]
                    if itemToBuild then
                        placeableBy = {item = itemToBuild, count = 1}
                    end
                end
                if not placeableBy then
                    goto __continue19
                end
                local flags = __TS__ArrayFilter(
                    prototype.flags,
                    function(____, flag) return flagsToTransfer[flag] ~= nil end
                )
                flags[#flags + 1] = "hidden"
                flags[#flags + 1] = "not-on-map"
                local selectionBox = prototype.selection_box or ({{-0.5, -0.5}, {0.5, 0.5}})
                if rollingStockTypes[____type] ~= nil then
                    selectionBox = BBox.expand(
                        BBox.normalize(selectionBox),
                        0.3
                    ):asArray()
                end
                local isRail = railPrototypes[____type] ~= nil
                if isRail then
                    previews[#previews + 1] = createRailPreview(prototype, placeableBy, flags)
                else
                    local color = prototype.friendly_map_color or prototype.map_color or utilityConstants.chart.default_friendly_color_by_type[____type] or utilityConstants.chart.default_friendly_color
                    local preview = {
                        name = "bp100:preview-entity-" .. name,
                        type = "simple-entity-with-owner",
                        localised_name = {"bp100.preview-entity", {"entity-name." .. name}},
                        icons = prototype.icons,
                        icon_size = prototype.icon_size,
                        icon_mipmaps = prototype.icon_mipmaps,
                        icon = prototype.icon,
                        selection_box = selectionBox,
                        collision_box = prototype.collision_box,
                        tile_height = prototype.tile_height,
                        tile_width = prototype.tile_width,
                        collision_mask = {},
                        open_sound = prototype.open_sound,
                        close_sound = prototype.close_sound,
                        picture = ____exports.createWhiteSprite(selectionBox, color),
                        flags = flags,
                        placeable_by = placeableBy,
                        render_layer = "ground-patch-higher2",
                        secondary_draw_order = 100,
                        subgroup = "bp100:preview-entity",
                        create_ghost_on_death = false
                    }
                    previews[#previews + 1] = preview
                end
            end
            ::__continue19::
        end
    end
    ::__continue17::
end
data:extend(previews)
local previewNames = __TS__ArrayMap(
    previews,
    function(____, preview) return preview.name end
)
local ____array_1 = __TS__SparseArrayNew(table.unpack(buildableNames))
__TS__SparseArrayPush(
    ____array_1,
    table.unpack(previewNames)
)
local entityOrPreviewNames = {__TS__SparseArraySpread(____array_1)}
local cleanupTool = data.raw["selection-tool"]["bp100:cleanup-tool"]
local stageMoveTool = data.raw["selection-tool"]["bp100:stage-move-tool"]
local stagedCopyTool = data.raw["selection-tool"]["bp100:staged-copy-tool"]
local stagedCutTool = data.raw["selection-tool"]["bp100:staged-cut-tool"]
local forceDeleteTool = data.raw["selection-tool"]["bp100:force-delete-tool"]
cleanupTool.reverse_entity_filters = previewNames
local ____previewNames_2 = previewNames
cleanupTool.alt_entity_filters = ____previewNames_2
cleanupTool.entity_filters = ____previewNames_2
stageMoveTool.entity_filters = buildableNames
stageMoveTool.reverse_entity_filters = entityOrPreviewNames
stageMoveTool.alt_entity_filters = entityOrPreviewNames
stageMoveTool.alt_reverse_entity_filters = previewNames
stagedCopyTool.alt_entity_filters = buildableNames
stagedCopyTool.entity_filters = buildableNames
stagedCutTool.alt_entity_filters = buildableNames
stagedCutTool.entity_filters = buildableNames
forceDeleteTool.entity_filters = buildableNames
forceDeleteTool.alt_entity_filters = entityOrPreviewNames
local twoDirectionOnlyTanks = {}
for name, tank in pairs(data.raw["storage-tank"]) do
    if tank.two_direction_only then
        twoDirectionOnlyTanks[#twoDirectionOnlyTanks + 1] = name
    end
end
local prototypeInfo = {twoDirectionOnlyTanks = twoDirectionOnlyTanks}
data:extend({{
    name = "bp100:passed-prototype-info",
    type = "item",
    localised_description = serpent.dump(prototypeInfo),
    flags = {"hidden"},
    icon = empty_sprite().filename,
    icon_size = 1,
    stack_size = 1
}})
return ____exports
