local ____lualib = require("lualib_bundle")
local __TS__Class = ____lualib.__TS__Class
local __TS__StringSubstring = ____lualib.__TS__StringSubstring
local __TS__DecorateLegacy = ____lualib.__TS__DecorateLegacy
local __TS__New = ____lualib.__TS__New
local ____exports = {}
local ____util = require("util")
local oppositedirection = ____util.oppositedirection
local ____lib = require("lib.index")
local isEmpty = ____lib.isEmpty
local RegisterClass = ____lib.RegisterClass
local ____geometry = require("lib.geometry.index")
local BBox = ____geometry.BBox
local ____entity_2Dprototype_2Dinfo = require("entity.entity-prototype-info")
local isPreviewEntity = ____entity_2Dprototype_2Dinfo.isPreviewEntity
local isRollingStockType = ____entity_2Dprototype_2Dinfo.isRollingStockType
local OnEntityPrototypesLoaded = ____entity_2Dprototype_2Dinfo.OnEntityPrototypesLoaded
local rollingStockTypes = ____entity_2Dprototype_2Dinfo.rollingStockTypes
local ____map2d = require("entity.map2d")
local _migrateMap2DToLinkedList = ____map2d._migrateMap2DToLinkedList
local newMap2D = ____map2d.newMap2D
local ____registration = require("entity.registration")
local getRegisteredProjectEntity = ____registration.getRegisteredProjectEntity
local ____underground_2Dbelt = require("entity.underground-belt")
local getUndergroundDirection = ____underground_2Dbelt.getUndergroundDirection
local nameToType
local nameToCategory
local pasteCompatibleRotations
OnEntityPrototypesLoaded:addListener(function(i)
    local ____i_0 = i
    nameToType = ____i_0.nameToType
    nameToCategory = ____i_0.nameToCategory
    pasteCompatibleRotations = ____i_0.rotationTypes
end)
local ProjectContentImpl = __TS__Class()
ProjectContentImpl.name = "ProjectContentImpl"
function ProjectContentImpl.prototype.____constructor(self)
    self.byPosition = newMap2D()
    self.entities = {}
end
function ProjectContentImpl.prototype.has(self, entity)
    return self.entities[entity] ~= nil
end
function ProjectContentImpl.prototype.findCompatibleByProps(self, entityName, position, direction, stage)
    local ____position_1 = position
    local x = ____position_1.x
    local y = ____position_1.y
    local cur = self.byPosition:get(x, y)
    if not cur then
        return
    end
    local category = nameToCategory[entityName]
    local candidate = nil
    while cur ~= nil do
        local name = cur.firstValue.name
        if (direction == nil or direction == cur.direction) and (cur.lastStage == nil or cur.lastStage >= stage) and (name == entityName or category and nameToCategory[name] == category) and (candidate == nil or cur.firstStage < candidate.firstStage) then
            candidate = cur
        end
        cur = cur._next
    end
    return candidate
end
function ProjectContentImpl.prototype.findCompatibleWithLuaEntity(self, entity, previousDirection, stage)
    local ____type = entity.type
    if ____type == "underground-belt" then
        local found = self:findCompatibleByProps(____type, entity.position, nil, stage)
        if found and getUndergroundDirection(found.direction, found.firstValue.type) == getUndergroundDirection(entity.direction, entity.belt_to_ground_type) then
            return found
        end
        return nil
    end
    if rollingStockTypes[____type] ~= nil then
        if entity.object_name == "LuaEntity" then
            local registered = getRegisteredProjectEntity(entity)
            if registered and self.entities[registered] ~= nil then
                return registered
            end
        end
        return nil
    end
    local name = entity.name
    local pasteRotatableType = pasteCompatibleRotations[name]
    if pasteRotatableType == nil then
        return self:findCompatibleByProps(name, entity.position, previousDirection or entity.direction, stage)
    end
    if pasteRotatableType == 1 then
        return self:findCompatibleByProps(name, entity.position, nil, stage)
    end
    if pasteRotatableType == 0 then
        local direction = previousDirection or entity.direction
        local position = entity.position
        if direction % 2 == 1 then
            return self:findCompatibleByProps(name, position, direction, stage)
        end
        return self:findCompatibleByProps(name, position, direction, stage) or self:findCompatibleByProps(
            name,
            position,
            oppositedirection(direction),
            stage
        )
    end
end
function ProjectContentImpl.prototype.findCompatibleWithExistingEntity(self, entity, stage)
    local name = entity.firstValue.name
    return self:findCompatibleWithLuaEntity(
        {
            name = name,
            type = nameToType[name] or "unknown",
            position = entity.position,
            direction = entity.direction,
            belt_to_ground_type = entity:isUndergroundBelt() and entity.firstValue.type or nil
        },
        nil,
        stage
    )
end
function ProjectContentImpl.prototype.findCompatibleFromPreview(self, previewEntity, stage)
    local actualName = __TS__StringSubstring(previewEntity.name, #"bp100:preview-entity-")
    local direction = isRollingStockType(actualName) and 0 or previewEntity.direction
    return self:findCompatibleByProps(actualName, previewEntity.position, direction, stage)
end
function ProjectContentImpl.prototype.findCompatibleFromLuaEntityOrPreview(self, entity, stage)
    if isPreviewEntity(entity) then
        return self:findCompatibleFromPreview(entity, stage)
    end
    return self:findCompatibleWithLuaEntity(entity, nil, stage)
end
function ProjectContentImpl.prototype.findExact(self, entity, position, stage)
    local cur = self.byPosition:get(position.x, position.y)
    while cur ~= nil do
        if cur:getWorldOrPreviewEntity(stage) == entity then
            return cur
        end
        cur = cur._next
    end
    return nil
end
function ProjectContentImpl.prototype.countNumEntities(self)
    return table_size(self.entities)
end
function ProjectContentImpl.prototype.iterateAllEntities(self)
    return self.entities
end
function ProjectContentImpl.prototype.computeBoundingBox(self)
    if isEmpty(self.entities) then
        return nil
    end
    local minX = math.huge
    local minY = math.huge
    local maxX = -math.huge
    local maxY = -math.huge
    for entity in pairs(self.entities) do
        local ____entity_position_2 = entity.position
        local x = ____entity_position_2.x
        local y = ____entity_position_2.y
        if x < minX then
            minX = x
        end
        if y < minY then
            minY = y
        end
        if x > maxX then
            maxX = x
        end
        if y > maxY then
            maxY = y
        end
    end
    return BBox.expand(
        BBox.coords(minX, minY, maxX, maxY),
        20
    )
end
function ProjectContentImpl.prototype.add(self, entity)
    local entities = self.entities
    if entities[entity] ~= nil then
        return
    end
    entities[entity] = true
    local ____entity_position_3 = entity.position
    local x = ____entity_position_3.x
    local y = ____entity_position_3.y
    self.byPosition:add(x, y, entity)
    entity:addOrPruneIngoingConnections(entities)
end
function ProjectContentImpl.prototype.delete(self, entity)
    local entities = self.entities
    if not (entities[entity] ~= nil) then
        return
    end
    entities[entity] = nil
    local ____entity_position_4 = entity.position
    local x = ____entity_position_4.x
    local y = ____entity_position_4.y
    self.byPosition:delete(x, y, entity)
    entity:removeIngoingConnections()
end
function ProjectContentImpl.prototype.changePosition(self, entity, position)
    if not (self.entities[entity] ~= nil) then
        return false
    end
    local ____entity_position_5 = entity.position
    local x = ____entity_position_5.x
    local y = ____entity_position_5.y
    local ____position_6 = position
    local newX = ____position_6.x
    local newY = ____position_6.y
    if x == newX and y == newY then
        return false
    end
    local byPosition = self.byPosition
    byPosition:delete(x, y, entity)
    entity:setPositionUnchecked(position)
    byPosition:add(newX, newY, entity)
    return true
end
function ProjectContentImpl.prototype.insertStage(self, stageNumber)
    for entity in pairs(self.entities) do
        entity:insertStage(stageNumber)
    end
end
function ProjectContentImpl.prototype.deleteStage(self, stageNumber)
    for entity in pairs(self.entities) do
        entity:deleteStage(stageNumber)
    end
end
function ProjectContentImpl.prototype.__tostring(self)
    return ("ProjectContent(" .. tostring(self:countNumEntities())) .. " entities)"
end
ProjectContentImpl = __TS__DecorateLegacy(
    {RegisterClass("EntityMap")},
    ProjectContentImpl
)
function ____exports._assertCorrect(content)
    local ____content_7 = content
    local entities = ____content_7.entities
    for entity in pairs(entities) do
        local points = entity.circuitConnections
        if points then
            for otherEntity, connections in pairs(points) do
                assert(entities[otherEntity] ~= nil)
                for connection in pairs(connections) do
                    local ____assert_12 = assert
                    local ____opt_10 = otherEntity.circuitConnections
                    local ____opt_8 = ____opt_10 and ____opt_10[entity]
                    ____assert_12(____opt_8 and ____opt_8[connection] ~= nil)
                end
            end
        end
        if entity.cableConnections then
            for otherEntity in pairs(entity.cableConnections) do
                assert(entities[otherEntity] ~= nil)
                local ____assert_15 = assert
                local ____opt_13 = otherEntity.cableConnections
                ____assert_15(____opt_13 and ____opt_13[entity] ~= nil)
            end
        end
    end
end
function ____exports.newProjectContent()
    return __TS__New(ProjectContentImpl)
end
function ____exports._migrateProjectContent_0_18_0(content)
    _migrateMap2DToLinkedList(content.byPosition)
end
function ____exports._migrateWireConnections(content)
    local ____content_16 = content
    local circuitConnections = ____content_16.circuitConnections
    local cableConnections = ____content_16.cableConnections
    if cableConnections then
        for entity, connections in pairs(cableConnections) do
            entity.cableConnections = connections
        end
    end
    if circuitConnections then
        for entity, connections in pairs(circuitConnections) do
            entity.circuitConnections = connections
        end
    end
    content.circuitConnections = nil
    content.cableConnections = nil
end
return ____exports
