local ____lualib = require("lualib_bundle")
local __TS__StringStartsWith = ____lualib.__TS__StringStartsWith
local ____exports = {}
local getPassedPrototypeInfo, computeEntityPrototypeInfo
local ____util = require("util")
local list_to_map = ____util.list_to_map
local merge = ____util.merge
local ____lib = require("lib.index")
local Events = ____lib.Events
local globalEvent = ____lib.globalEvent
local ____geometry = require("lib.geometry.index")
local BBox = ____geometry.BBox
function getPassedPrototypeInfo()
    local fakeItem = game.item_prototypes["bp100:passed-prototype-info"]
    local success, value = serpent.load(fakeItem.localised_description)
    if not success then
        error("Failed to parse passed prototype info.")
    end
    return value
end
function computeEntityPrototypeInfo()
    log("Processing blueprint-able entity prototypes")
    local compatibleTypes = {["logistic-container"] = "container", ["rail-chain-signal"] = "rail-signal", ["storage-tank"] = "pipe"}
    local ignoreFastReplaceGroup = {["transport-belt"] = true, ["underground-belt"] = true, splitter = true}
    local nameToCategory = {}
    local categories = {}
    local rotationTypes = {}
    local selectionBoxes = {}
    local passedPrototypeInfo = getPassedPrototypeInfo()
    local twoDirectionTanks = list_to_map(passedPrototypeInfo.twoDirectionOnlyTanks)
    local pasteRotatableEntityTypes = {["assembling-machine"] = true, boiler = true, generator = true}
    local function getCategory(prototype)
        local ____prototype_2 = prototype
        local fast_replaceable_group = ____prototype_2.fast_replaceable_group
        local ____type = ____prototype_2.type
        local collision_box = ____prototype_2.collision_box
        local actualFastReplaceGroup = ignoreFastReplaceGroup[____type] ~= nil and "" or fast_replaceable_group
        if actualFastReplaceGroup == nil then
            return
        end
        local actualType = compatibleTypes[____type] or ____type
        local ____collision_box_left_top_3 = collision_box.left_top
        local lx = ____collision_box_left_top_3.x
        local ly = ____collision_box_left_top_3.y
        local ____collision_box_right_bottom_4 = collision_box.right_bottom
        local rx = ____collision_box_right_bottom_4.x
        local ry = ____collision_box_right_bottom_4.y
        return table.concat({
            actualType,
            actualFastReplaceGroup,
            lx,
            ly,
            rx,
            ry
        }, "|")
    end
    local function getPasteCompatibleRotation(prototype)
        if not prototype.supports_direction or prototype.has_flag("not-rotatable") then
            return 1
        end
        local ____type = prototype.type
        if ____type == "straight-rail" then
            return 0
        end
        if pasteRotatableEntityTypes[____type] ~= nil then
            local collisionBox = prototype.collision_box
            if BBox.isCenteredSquare(collisionBox) then
                return 1
            end
            if BBox.isCenteredRectangle(collisionBox) then
                return 0
            end
        end
    end
    local nameToType = {}
    for name, prototype in pairs(game.get_filtered_entity_prototypes({{filter = "blueprintable"}})) do
        local categoryName = getCategory(prototype)
        if categoryName ~= nil then
            nameToCategory[name] = categoryName
            local category = categories[categoryName]
            if category == nil then
                category = {}
                categories[categoryName] = category
            end
            category[#category + 1] = name
        end
        local pasteRotationType = getPasteCompatibleRotation(prototype)
        if pasteRotationType ~= nil then
            rotationTypes[name] = pasteRotationType
        end
        selectionBoxes[name] = BBox.from(prototype.selection_box)
        nameToType[name] = prototype.type
    end
    for categoryName, category in pairs(categories) do
        if table_size(category) <= 1 then
            categories[categoryName] = nil
            for ____, name in ipairs(category) do
                nameToCategory[name] = nil
            end
        end
    end
    return {
        nameToCategory = nameToCategory,
        categories = categories,
        rotationTypes = rotationTypes,
        selectionBoxes = selectionBoxes,
        nameToType = nameToType,
        twoDirectionTanks = twoDirectionTanks
    }
end
local entityPrototypeInfo
____exports.OnEntityPrototypesLoaded = globalEvent()
Events.on_configuration_changed(function()
    local ____computeEntityPrototypeInfo_result_0 = computeEntityPrototypeInfo()
    global.entityPrototypeInfo = ____computeEntityPrototypeInfo_result_0
    entityPrototypeInfo = ____computeEntityPrototypeInfo_result_0
    ____exports.OnEntityPrototypesLoaded:raise(entityPrototypeInfo)
end)
Events.on_init(function()
    local ____computeEntityPrototypeInfo_result_1 = computeEntityPrototypeInfo()
    global.entityPrototypeInfo = ____computeEntityPrototypeInfo_result_1
    entityPrototypeInfo = ____computeEntityPrototypeInfo_result_1
    ____exports.OnEntityPrototypesLoaded:raise(entityPrototypeInfo)
end)
Events.on_load(function()
    entityPrototypeInfo = global.entityPrototypeInfo
    if entityPrototypeInfo ~= nil then
        ____exports.OnEntityPrototypesLoaded:raise(entityPrototypeInfo)
    end
end)
local rollingStockTypes = {["artillery-wagon"] = true, ["cargo-wagon"] = true, ["fluid-wagon"] = true, locomotive = true}
____exports.allowOverlapDifferentDirection = merge({{["straight-rail"] = true, ["curved-rail"] = true}, rollingStockTypes})
function ____exports.getEntityPrototypeInfo()
    return entityPrototypeInfo
end
local nameToCategory
local nameToType
local categories
local rotationTypes
____exports.OnEntityPrototypesLoaded:addListener(function(info)
    local ____info_5 = info
    nameToCategory = ____info_5.nameToCategory
    categories = ____info_5.categories
    rotationTypes = ____info_5.rotationTypes
    nameToType = ____info_5.nameToType
end)
function ____exports.areUpgradeableTypes(a, b)
    if a == b then
        return true
    end
    local aCategory = nameToCategory[a]
    if aCategory == nil then
        return false
    end
    return aCategory == nameToCategory[b]
end
function ____exports.getCompatibleNames(entityName)
    local category = nameToCategory[entityName]
    if category == nil then
        return
    end
    return categories[category]
end
--- For straight rails, paste rotation only applies to non-diagonal rails.
function ____exports.getPrototypeRotationType(entityName)
    return rotationTypes[entityName]
end
function ____exports.isRollingStockType(entityName)
    local ____type = nameToType[entityName]
    return ____type ~= nil and rollingStockTypes[____type] ~= nil
end
____exports.rollingStockTypes = rollingStockTypes
function ____exports.isPreviewEntity(entity)
    local ____type = entity.type
    return (____type == "simple-entity-with-owner" or ____type == "rail-remnants") and __TS__StringStartsWith(entity.name, "bp100:preview-entity-")
end
return ____exports
