local ____lualib = require("lualib_bundle")
local __TS__ArrayFilter = ____lualib.__TS__ArrayFilter
local __TS__ArrayMap = ____lualib.__TS__ArrayMap
local __TS__ArrayEvery = ____lualib.__TS__ArrayEvery
local ____exports = {}
local ____geometry = require("lib.geometry.index")
local BBox = ____geometry.BBox
local Pos = ____geometry.Pos
local function canMoveEntity(entity, position, direction)
    return entity.surface.can_place_entity({
        name = entity.name == "entity-ghost" and entity.ghost_name or entity.name,
        position = position,
        direction = direction,
        force = entity.force,
        build_check_type = defines.build_check_type.blueprint_ghost,
        inner_name = entity.name == "entity-ghost" and entity.ghost_name or nil
    })
end
local function canWiresReach(entity)
    local wires = entity.type == "electric-pole" and entity.neighbours or entity.circuit_connected_entities
    if not wires then
        return true
    end
    for ____, entities in pairs(wires) do
        for ____, e in ipairs(entities) do
            if not entity.can_wires_reach(e) then
                return false
            end
        end
    end
    return true
end
local function updateMovedEntity(oldPosition, entity)
    local proxy = entity.surface.find_entity("item-request-proxy", oldPosition)
    if proxy and proxy.valid then
        proxy.teleport(entity.position)
    end
    entity.update_connections()
    local updateableEntities = entity.surface.find_entities_filtered({
        area = BBox.expand(entity.bounding_box, 32),
        force = entity.force
    })
    for ____, e in ipairs(updateableEntities) do
        e.update_connections()
    end
end
local function tryMoveOutOfTheWay(entity)
    local retries = 5
    local outOfTheWay = Pos.plus(entity.position, {x = 20, y = 20})
    while not entity.teleport(outOfTheWay) do
        if retries <= 1 then
            return false
        end
        outOfTheWay.x = outOfTheWay.x + retries
        outOfTheWay.y = outOfTheWay.y + retries
        retries = retries - 1
    end
    return true
end
--- Moves all entities back if not successful.
function ____exports.tryDollyAllEntities(entities, position, direction)
    entities = __TS__ArrayFilter(
        entities,
        function(____, e) return e.valid and not (Pos.equals(e.position, position) and e.direction == direction) end
    )
    local oldPositions = __TS__ArrayMap(
        entities,
        function(____, e) return e.position end
    )
    local oldDirections = __TS__ArrayMap(
        entities,
        function(____, e) return e.direction end
    )
    local function revertTeleport()
        for i = 1, #entities do
            local entity = entities[i]
            entity.direction = oldDirections[i]
            entity.teleport(oldPositions[i])
        end
    end
    if not __TS__ArrayEvery(
        entities,
        function(____, e) return tryMoveOutOfTheWay(e) end
    ) then
        revertTeleport()
        return "could-not-teleport"
    end
    if not __TS__ArrayEvery(
        entities,
        function(____, e) return canMoveEntity(e, position, direction) end
    ) then
        revertTeleport()
        return "overlap"
    end
    local teleportError = false
    for ____, entity in ipairs(entities) do
        entity.direction = direction
        if not entity.teleport(position) then
            teleportError = true
            break
        end
    end
    if teleportError or not __TS__ArrayEvery(
        entities,
        function(____, e) return canWiresReach(e) end
    ) then
        revertTeleport()
        return teleportError and "could-not-teleport" or "wires-cannot-reach"
    end
    for i = 1, #entities do
        updateMovedEntity(oldPositions[i], entities[i])
    end
    return "success"
end
function ____exports.forceDollyEntity(entity, position, direction)
    local oldPosition = entity.position
    local oldDirection = entity.direction
    entity.direction = direction
    if not entity.teleport(position) then
        entity.direction = oldDirection
        return false
    end
    updateMovedEntity(oldPosition, entity)
    return true
end
return ____exports
