local ____lualib = require("lualib_bundle")
local __TS__ArrayMap = ____lualib.__TS__ArrayMap
local ____exports = {}
local ____blueprint_2Dstage_2Dinfo = require("copy-paste.blueprint-stage-info")
local fromBpStageDiffs = ____blueprint_2Dstage_2Dinfo.fromBpStageDiffs
local ____empty_2Dcontrol_2Dbehavior = require("entity.empty-control-behavior")
local maybeSetEmptyControlBehavior = ____empty_2Dcontrol_2Dbehavior.maybeSetEmptyControlBehavior
local ____entity_2Dprototype_2Dinfo = require("entity.entity-prototype-info")
local areUpgradeableTypes = ____entity_2Dprototype_2Dinfo.areUpgradeableTypes
local ____ProjectEntity = require("entity.ProjectEntity")
local createProjectEntityNoCopy = ____ProjectEntity.createProjectEntityNoCopy
local ____save_2Dload = require("entity.save-load")
local canBeAnyDirection = ____save_2Dload.canBeAnyDirection
local copyKnownValue = ____save_2Dload.copyKnownValue
local forceFlipUnderground = ____save_2Dload.forceFlipUnderground
local saveEntity = ____save_2Dload.saveEntity
local ____underground_2Dbelt = require("entity.underground-belt")
local findUndergroundPair = ____underground_2Dbelt.findUndergroundPair
local ____wires = require("entity.wires")
local saveWireConnections = ____wires.saveWireConnections
local ____geometry = require("lib.geometry.index")
local Pos = ____geometry.Pos
local min = math.min
---
-- @noSelf
function ____exports.ProjectUpdates(project, worldEntityUpdates)
    local fixNewUndergroundBelt, createNewProjectEntity, addNewEntity, shouldMakeSettingsRemnant, deleteEntityOrCreateSettingsRemnant, forceDeleteEntity, readdDeletedEntity, tryReviveSettingsRemnant, applyValueFromWorld, tryUpdateEntityFromWorld, tryRotateEntityFromWorld, checkUpgradeType, tryUpgradeEntityFromWorld, handleUpdate, updatePair, handleUndergroundFlippedBack, doUndergroundBeltUpdate, handleUndergroundBeltUpdate, maybeApplyEmptyControlBehavior, updateWiresFromWorld, handleUndergroundBeltValueSet, setValueFromStagedInfo, firstStageChangeWillIntersect, lastStageChangeWillIntersect, checkCanSetFirstStage, trySetFirstStage, checkCanSetLastStage, trySetLastStage, resetProp, movePropDown, resetAllProps, moveAllPropsDown, resetTrain, setTrainLocationToCurrent, content, deleteWorldEntities, makeSettingsRemnant, rebuildWorldEntityAtStage, refreshWorldEntityAtStage, resetUnderground, reviveSettingsRemnant, updateNewWorldEntitiesWithoutWires, updateWireConnections, updateWorldEntities, updateWorldEntitiesOnLastStageChanged, updateAllHighlights
    function fixNewUndergroundBelt(projectEntity, entity, stage, knownValue)
        if entity.type ~= "underground-belt" then
            return
        end
        if knownValue then
            projectEntity:setTypeProperty(entity.belt_to_ground_type)
        end
        local pair = findUndergroundPair(content, projectEntity, stage)
        if pair then
            local expectedType = pair.firstValue.type == "output" and "input" or "output"
            if expectedType ~= projectEntity.firstValue.type then
                projectEntity:setTypeProperty(expectedType)
                projectEntity.direction = pair.direction
            end
        end
    end
    function createNewProjectEntity(entity, stage, knownValue)
        local ____opt_1 = knownValue and knownValue.tags
        local stageInfo = ____opt_1 and ____opt_1.bp100
        if not stageInfo then
            local saved = saveEntity(entity, knownValue)
            if not saved then
                return nil
            end
            return createProjectEntityNoCopy(saved, entity.position, entity.direction, stage)
        end
        local projectEntity = createProjectEntityNoCopy(
            stageInfo.firstValue or copyKnownValue(knownValue),
            entity.position,
            entity.direction,
            stageInfo.firstStage
        )
        projectEntity:setLastStageUnchecked(stageInfo.lastStage)
        local diffs = stageInfo.stageDiffs
        if diffs then
            projectEntity:setStageDiffs(fromBpStageDiffs(diffs))
        end
        return projectEntity
    end
    function addNewEntity(entity, stage, knownValue)
        local projectEntity = createNewProjectEntity(entity, stage, knownValue)
        if not projectEntity then
            return nil
        end
        projectEntity:replaceWorldEntity(stage, entity)
        content:add(projectEntity)
        fixNewUndergroundBelt(projectEntity, entity, stage, knownValue)
        updateNewWorldEntitiesWithoutWires(projectEntity)
        local hasDiff, ____, additionalToUpdate = saveWireConnections(
            content,
            projectEntity,
            stage,
            project:lastStageFor(projectEntity)
        )
        if hasDiff then
            updateWireConnections(projectEntity)
            if additionalToUpdate then
                for ____, otherEntity in ipairs(additionalToUpdate) do
                    updateWireConnections(otherEntity)
                end
            end
        end
        if projectEntity.stageDiffs then
            updateAllHighlights(projectEntity)
        end
        return projectEntity
    end
    function shouldMakeSettingsRemnant(entity)
        if entity:hasStageDiff() then
            return true
        end
        local connections = entity.circuitConnections
        if not connections then
            return false
        end
        local stage = entity.firstStage
        for otherEntity in pairs(connections) do
            if otherEntity:getWorldEntity(stage) == nil then
                return true
            end
        end
        return false
    end
    function deleteEntityOrCreateSettingsRemnant(entity)
        if shouldMakeSettingsRemnant(entity) then
            entity.isSettingsRemnant = true
            makeSettingsRemnant(entity)
        else
            content:delete(entity)
            deleteWorldEntities(entity)
        end
    end
    function forceDeleteEntity(entity)
        content:delete(entity)
        deleteWorldEntities(entity)
    end
    function readdDeletedEntity(entity)
        content:add(entity)
        updateWorldEntities(entity, 1)
    end
    function tryReviveSettingsRemnant(entity, stage)
        if not entity.isSettingsRemnant then
            return "no-change"
        end
        local result = checkCanSetFirstStage(entity, stage)
        if result == "updated" or result == "no-change" then
            entity:setFirstStageUnchecked(stage)
            entity.isSettingsRemnant = nil
            reviveSettingsRemnant(entity)
        end
        return result
    end
    function applyValueFromWorld(stage, entity, entitySource, knownValue)
        local newValue = saveEntity(entitySource, knownValue)
        return newValue ~= nil and entity:adjustValueAtStage(stage, newValue)
    end
    function tryUpdateEntityFromWorld(entity, stage, knownValue)
        local ____opt_5 = knownValue and knownValue.tags
        if ____opt_5 and ____opt_5.bp100 then
            error("Should call setValueFromStagedInfo instead")
        end
        local entitySource = entity:getWorldEntity(stage)
        if not entitySource then
            return "no-change"
        end
        return handleUpdate(
            entity,
            entitySource,
            stage,
            entitySource.direction,
            nil,
            true,
            knownValue
        )
    end
    function tryRotateEntityFromWorld(entity, stage)
        local entitySource = entity:getWorldEntity(stage)
        if not entitySource then
            return "no-change"
        end
        return handleUpdate(
            entity,
            entitySource,
            stage,
            entitySource.direction,
            nil,
            false,
            nil
        )
    end
    function checkUpgradeType(existing, upgradeType)
        if not areUpgradeableTypes(existing.firstValue.name, upgradeType) then
            error(((" incompatible upgrade from " .. existing.firstValue.name) .. " to ") .. upgradeType)
        end
    end
    function tryUpgradeEntityFromWorld(entity, stage)
        local entitySource = entity:getWorldEntity(stage)
        if not entitySource then
            return "no-change"
        end
        local ____opt_9 = entitySource.get_upgrade_target()
        local overrideUpgradeTarget = ____opt_9 and ____opt_9.name
        return handleUpdate(
            entity,
            entitySource,
            stage,
            entitySource.get_upgrade_direction(),
            overrideUpgradeTarget,
            false,
            nil
        )
    end
    function handleUpdate(entity, entitySource, stage, targetDirection, targetUpgrade, getBpValue, knownBpValue)
        if entity:isUndergroundBelt() then
            return handleUndergroundBeltUpdate(
                entity,
                entitySource,
                stage,
                targetDirection,
                targetUpgrade or knownBpValue and knownBpValue.name or entitySource.name
            )
        end
        local rotated = targetDirection and targetDirection ~= entity.direction and not canBeAnyDirection(entitySource)
        if rotated then
            local rotateAllowed = stage == entity.firstStage
            if rotateAllowed then
                entity.direction = targetDirection
                local entityType = entitySource.type
                if entityType == "loader" or entityType == "loader-1x1" then
                    entity:setTypeProperty(entitySource.loader_type)
                elseif entityType == "inserter" then
                    if entity.firstValue.pickup_position then
                        entity:setPickupPosition(Pos.minus(entitySource.pickup_position, entitySource.position))
                    end
                    if entity.firstValue.drop_position then
                        entity:setDropPosition(Pos.minus(entitySource.drop_position, entitySource.position))
                    end
                end
            else
                refreshWorldEntityAtStage(entity, stage)
                return "cannot-rotate"
            end
        end
        local hasDiff = false
        if getBpValue and applyValueFromWorld(stage, entity, entitySource, knownBpValue) then
            hasDiff = true
        elseif targetUpgrade then
            checkUpgradeType(entity, targetUpgrade)
            if entity:applyUpgradeAtStage(stage, targetUpgrade) then
                hasDiff = true
            end
        end
        if rotated or hasDiff then
            updateWorldEntities(entity, stage)
            return "updated"
        end
        return "no-change"
    end
    function updatePair(entity1, entity1Stage, entity2, entity2Stage)
        updateWorldEntities(entity1, entity1Stage, false)
        updateWorldEntities(entity2, entity2Stage, false)
        updateAllHighlights(entity1)
        updateAllHighlights(entity2)
    end
    function handleUndergroundFlippedBack(entity, worldEntity, stage, targetDirection, pair)
        if not pair then
            updateWorldEntities(entity, stage)
            return "no-change"
        end
        if pair.direction == targetDirection then
            updatePair(entity, entity.firstStage, pair, pair.firstStage)
            return "no-change"
        end
        local rotateAllowed = stage == entity.firstStage or pair.firstStage == stage
        if not rotateAllowed then
            forceFlipUnderground(worldEntity)
            return "cannot-rotate"
        end
        pair.direction = worldEntity.direction
        local oppositeType = worldEntity.belt_to_ground_type == "input" and "output" or "input"
        pair:setTypeProperty(oppositeType)
        updatePair(entity, entity.firstStage, pair, pair.firstStage)
        return "updated"
    end
    function doUndergroundBeltUpdate(entity, worldEntity, pair, stage, targetDirection, targetUpgrade)
        local rotated = targetDirection and targetDirection ~= entity.direction
        local oldName = entity:getNameAtStage(stage)
        local upgraded = targetUpgrade ~= oldName
        if not rotated and not upgraded then
            if not targetDirection then
                return "no-change"
            end
            return handleUndergroundFlippedBack(
                entity,
                worldEntity,
                stage,
                targetDirection,
                pair
            )
        end
        local isSelfOrPairFirstStage = stage == entity.firstStage or pair and pair.firstStage == stage
        if rotated then
            local rotateAllowed = isSelfOrPairFirstStage
            if not rotateAllowed then
                resetUnderground(entity, stage)
                return "cannot-rotate"
            end
            entity.direction = targetDirection
            local oldType = entity.firstValue.type
            local newType = oldType == "input" and "output" or "input"
            entity:setTypeProperty(newType)
            if pair then
                pair.direction = targetDirection
                pair:setTypeProperty(oldType)
            end
        end
        local applyStage = isSelfOrPairFirstStage and entity.firstStage or stage
        local pairApplyStage = pair and isSelfOrPairFirstStage and pair.firstStage or stage
        local cannotUpgradeChangedPair = false
        if upgraded then
            entity:applyUpgradeAtStage(applyStage, targetUpgrade)
            if pair ~= nil then
                pair:applyUpgradeAtStage(pairApplyStage, targetUpgrade)
            end
            local newPair = findUndergroundPair(content, entity, stage)
            cannotUpgradeChangedPair = newPair ~= pair
            if cannotUpgradeChangedPair then
                entity:applyUpgradeAtStage(stage, oldName)
                if pair ~= nil then
                    pair:applyUpgradeAtStage(pairApplyStage, oldName)
                end
            end
        end
        if cannotUpgradeChangedPair and not rotated then
            refreshWorldEntityAtStage(entity, stage)
            if pair then
                refreshWorldEntityAtStage(pair, stage)
            end
        elseif not pair then
            updateWorldEntities(entity, applyStage)
        else
            updatePair(entity, applyStage, pair, pairApplyStage)
        end
        return cannotUpgradeChangedPair and "cannot-upgrade-changed-pair" or "updated"
    end
    function handleUndergroundBeltUpdate(entity, worldEntity, stage, targetDirection, targetUpgrade)
        local pair = findUndergroundPair(content, entity, stage)
        local updateResult = doUndergroundBeltUpdate(
            entity,
            worldEntity,
            pair,
            stage,
            targetDirection,
            targetUpgrade
        )
        local newWorldEntity = entity:getWorldEntity(stage)
        if newWorldEntity then
            local worldPair = newWorldEntity.neighbours
            if worldPair and (not pair or pair:getWorldEntity(stage) ~= worldPair) then
                local worldPairEntity = content:findCompatibleWithLuaEntity(worldPair, nil, stage)
                if worldPairEntity then
                    updateAllHighlights(worldPairEntity)
                end
            end
        end
        return updateResult
    end
    function maybeApplyEmptyControlBehavior(entity, stage)
        if not maybeSetEmptyControlBehavior(entity, stage) then
            return false
        end
        local luaEntity = entity:getWorldEntity(stage)
        if luaEntity then
            applyValueFromWorld(stage, entity, luaEntity)
        end
        return true
    end
    function updateWiresFromWorld(entity, stage)
        local connectionsChanged, maxConnectionsExceeded, additionalEntitiesToUpdate = saveWireConnections(content, entity, stage, stage)
        if not connectionsChanged then
            return "no-change"
        end
        local circuitConnections = entity.circuitConnections
        if circuitConnections then
            maybeApplyEmptyControlBehavior(entity, stage)
        end
        updateWorldEntities(entity, entity.firstStage)
        if circuitConnections then
            for otherEntity in pairs(circuitConnections) do
                if maybeApplyEmptyControlBehavior(otherEntity, stage) then
                    updateWorldEntities(otherEntity, otherEntity.firstStage)
                end
            end
        end
        if additionalEntitiesToUpdate then
            for ____, otherEntity in ipairs(additionalEntitiesToUpdate) do
                updateWireConnections(otherEntity)
            end
        end
        if maxConnectionsExceeded then
            return "max-connections-exceeded"
        end
        return "updated"
    end
    function handleUndergroundBeltValueSet(entity, oldStageDiffs, stageDiffs)
        local possiblyUpdatedStages = {}
        if oldStageDiffs then
            for stage in pairs(oldStageDiffs) do
                possiblyUpdatedStages[stage] = true
            end
        end
        if stageDiffs then
            for stage in pairs(stageDiffs) do
                possiblyUpdatedStages[stage] = true
            end
        end
        local ugPairs = {}
        for stage in pairs(possiblyUpdatedStages) do
            local pair = findUndergroundPair(content, entity, stage)
            if pair then
                ugPairs[pair] = true
            end
        end
        for pair in pairs(ugPairs) do
            updateAllHighlights(pair)
        end
    end
    function setValueFromStagedInfo(entity, value, info)
        local targetStage = info.firstStage
        if targetStage ~= entity.firstStage then
            local result = checkCanSetFirstStage(entity, targetStage)
            if result ~= "updated" then
                return result
            end
            entity:setFirstStageUnchecked(targetStage)
        end
        local lastStage = info.lastStage
        local oldLastStage = entity.lastStage
        if lastStage ~= oldLastStage then
            local result = checkCanSetLastStage(entity, lastStage)
            if result ~= "updated" then
                return result
            end
            entity:setLastStageUnchecked(lastStage)
            if lastStage ~= nil and (oldLastStage == nil or lastStage < oldLastStage) then
                updateWorldEntitiesOnLastStageChanged(entity, oldLastStage)
            end
        end
        local oldStageDiffs = entity.stageDiffs
        local firstValue = info.firstValue or copyKnownValue(value)
        entity:setFirstValue(firstValue)
        local stageDiffs = info.stageDiffs and fromBpStageDiffs(info.stageDiffs) or nil
        entity:setStageDiffs(stageDiffs)
        updateWorldEntities(entity, 1)
        if entity:isUndergroundBelt() then
            handleUndergroundBeltValueSet(entity, oldStageDiffs, stageDiffs)
        end
        return "updated"
    end
    function firstStageChangeWillIntersect(entity, newStage)
        if newStage >= entity.firstStage then
            return true
        end
        local foundBelow = content:findCompatibleWithExistingEntity(entity, newStage)
        return foundBelow == nil or foundBelow == entity
    end
    function lastStageChangeWillIntersect(entity, newStage)
        local ____entity_17 = entity
        local lastStage = ____entity_17.lastStage
        if lastStage == nil or newStage ~= nil and newStage < lastStage then
            return true
        end
        local foundAbove = content:findCompatibleWithExistingEntity(entity, lastStage + 1)
        return foundAbove == nil or newStage ~= nil and foundAbove.firstStage > newStage
    end
    function checkCanSetFirstStage(entity, stage)
        if entity.isSettingsRemnant or entity.firstStage == stage then
            return "no-change"
        end
        if entity.lastStage and stage > entity.lastStage then
            return "cannot-move-past-last-stage"
        end
        if not firstStageChangeWillIntersect(entity, stage) then
            return "intersects-another-entity"
        end
        return "updated"
    end
    function trySetFirstStage(entity, stage)
        local result = checkCanSetFirstStage(entity, stage)
        if result == "updated" then
            local stageToUpdate = min(entity.firstStage, stage)
            entity:setFirstStageUnchecked(stage)
            updateWorldEntities(entity, stageToUpdate)
        end
        return result
    end
    function checkCanSetLastStage(entity, stage)
        if entity.isSettingsRemnant then
            return "no-change"
        end
        local oldLastStage = entity.lastStage
        if oldLastStage == stage then
            return "no-change"
        end
        if stage ~= nil and stage < entity.firstStage then
            return "cannot-move-before-first-stage"
        end
        if not lastStageChangeWillIntersect(entity, stage) then
            return "intersects-another-entity"
        end
        return "updated"
    end
    function trySetLastStage(entity, stage)
        if entity.isSettingsRemnant then
            return "no-change"
        end
        local result = checkCanSetLastStage(entity, stage)
        if result == "updated" then
            local oldLastStage = entity.lastStage
            entity:setLastStageUnchecked(stage)
            updateWorldEntitiesOnLastStageChanged(entity, oldLastStage)
        end
        return result
    end
    function resetProp(entity, stageNumber, prop)
        local moved = entity:resetProp(stageNumber, prop)
        if moved then
            updateWorldEntities(entity, stageNumber)
        end
        return moved
    end
    function movePropDown(entity, stageNumber, prop)
        local movedStage = entity:movePropDown(stageNumber, prop)
        if movedStage then
            updateWorldEntities(entity, movedStage)
            return true
        end
        return false
    end
    function resetAllProps(entity, stageNumber)
        local moved = entity:resetValue(stageNumber)
        if moved then
            updateWorldEntities(entity, stageNumber)
        end
        return moved
    end
    function moveAllPropsDown(entity, stageNumber)
        local movedStage = entity:moveValueDown(stageNumber)
        if movedStage then
            updateWorldEntities(entity, movedStage)
            return true
        end
        return false
    end
    function resetTrain(entity)
        local stage = entity.firstStage
        local luaEntity = entity:getWorldEntity(stage)
        if not luaEntity then
            refreshWorldEntityAtStage(entity, stage)
            return
        end
        local train = luaEntity.train
        if not train then
            return
        end
        local entities = train.carriages
        local projectEntities = __TS__ArrayMap(
            entities,
            function(____, e) return content:findCompatibleWithLuaEntity(e, nil, stage) end
        )
        for ____, entity in ipairs(projectEntities) do
            entity:destroyAllWorldOrPreviewEntities()
        end
        for ____, entity in ipairs(projectEntities) do
            rebuildWorldEntityAtStage(entity, stage)
        end
    end
    function setTrainLocationToCurrent(entity)
        local stage = entity.firstStage
        local luaEntity = entity:getWorldEntity(stage)
        if not luaEntity then
            return
        end
        local train = luaEntity.train
        if not train then
            return
        end
        local entities = train.carriages
        for ____, luaEntity in ipairs(entities) do
            local projectEntity = content:findCompatibleWithLuaEntity(luaEntity, nil, stage)
            if projectEntity then
                content:changePosition(projectEntity, luaEntity.position)
                rebuildWorldEntityAtStage(projectEntity, stage)
            else
                addNewEntity(luaEntity, stage)
            end
        end
    end
    content = project.content
    local ____worldEntityUpdates_0 = worldEntityUpdates
    deleteWorldEntities = ____worldEntityUpdates_0.deleteWorldEntities
    makeSettingsRemnant = ____worldEntityUpdates_0.makeSettingsRemnant
    rebuildWorldEntityAtStage = ____worldEntityUpdates_0.rebuildWorldEntityAtStage
    refreshWorldEntityAtStage = ____worldEntityUpdates_0.refreshWorldEntityAtStage
    resetUnderground = ____worldEntityUpdates_0.resetUnderground
    reviveSettingsRemnant = ____worldEntityUpdates_0.reviveSettingsRemnant
    updateNewWorldEntitiesWithoutWires = ____worldEntityUpdates_0.updateNewWorldEntitiesWithoutWires
    updateWireConnections = ____worldEntityUpdates_0.updateWireConnections
    updateWorldEntities = ____worldEntityUpdates_0.updateWorldEntities
    updateWorldEntitiesOnLastStageChanged = ____worldEntityUpdates_0.updateWorldEntitiesOnLastStageChanged
    updateAllHighlights = ____worldEntityUpdates_0.updateAllHighlights
    return {
        addNewEntity = addNewEntity,
        deleteEntityOrCreateSettingsRemnant = deleteEntityOrCreateSettingsRemnant,
        forceDeleteEntity = forceDeleteEntity,
        readdDeletedEntity = readdDeletedEntity,
        tryReviveSettingsRemnant = tryReviveSettingsRemnant,
        tryUpdateEntityFromWorld = tryUpdateEntityFromWorld,
        tryRotateEntityFromWorld = tryRotateEntityFromWorld,
        tryUpgradeEntityFromWorld = tryUpgradeEntityFromWorld,
        updateWiresFromWorld = updateWiresFromWorld,
        setValueFromStagedInfo = setValueFromStagedInfo,
        trySetFirstStage = trySetFirstStage,
        trySetLastStage = trySetLastStage,
        resetProp = resetProp,
        movePropDown = movePropDown,
        resetAllProps = resetAllProps,
        moveAllPropsDown = moveAllPropsDown,
        resetTrain = resetTrain,
        setTrainLocationToCurrent = setTrainLocationToCurrent
    }
end
return ____exports
