local ____lualib = require("lualib_bundle")
local __TS__ArrayEvery = ____lualib.__TS__ArrayEvery
local ____exports = {}
local registerUndo, registerUndoLater, _lastUndoIndex, FutureUndoTranslation
local ____lib = require("lib.index")
local Events = ____lib.Events
local onPlayerInitSince = ____lib.onPlayerInitSince
local ____migration = require("lib.migration")
local Migrations = ____migration.Migrations
function registerUndo(handlerName, player, data)
    local playerData = global.players[player.index]
    local index = playerData.nextUndoEntryIndex
    local entity = player.surface.create_entity({name = "entity-ghost", inner_name = "bp100:undo-reference", position = {index, 0}})
    if not entity or not player.mine_entity(entity, true) then
        error("Failed to make undo reference entity")
    end
    playerData.undoEntries[index] = {handlerName = handlerName, data = data}
    playerData.nextUndoEntryIndex = index % 100 + 1
    _lastUndoIndex = index
    return index
end
function registerUndoLater(handlerName, player, data)
    local id = player.request_translation(FutureUndoTranslation)
    if id then
        global.futureUndoData[id] = {handlerName = handlerName, data = data}
    end
end
local floor = math.floor
local function undoInit()
    global.futureUndoData = {}
    setmetatable(global.futureUndoData, {__mode = "v"})
end
Events.on_init(undoInit)
Migrations.fromAny(undoInit)
Events.on_load(function()
    if global.futureUndoData ~= nil then
        setmetatable(global.futureUndoData, {__mode = "v"})
    end
end)
onPlayerInitSince(
    "0.18.0",
    function(player)
        local playerData = global.players[player]
        playerData.undoEntries = {}
        playerData.nextUndoEntryIndex = 1
    end
)
local undoHandlers = {}
---
-- @noSelf
function ____exports.UndoHandler(name, fn)
    if undoHandlers[name] ~= nil then
        error("Undo handler already registered: " .. name)
    end
    undoHandlers[name] = fn
    return {
        register = function(player, data) return registerUndo(name, player, data) end,
        registerLater = function(player, data) return registerUndoLater(name, player, data) end,
        createAction = function(playerIndex, data) return {playerIndex = playerIndex, handlerName = name, data = data} end
    }
end
function ____exports.registerUndoAction(action)
    local player = game.get_player(action.playerIndex)
    if player then
        registerUndo(action.handlerName, player, action.data)
    end
end
function ____exports.registerUndoActionLater(action)
    local player = game.get_player(action.playerIndex)
    if player then
        registerUndoLater(action.handlerName, player, action.data)
    end
end
function ____exports.registerGroupUndoAction(actions)
    if #actions == 0 then
        return
    end
    if #actions == 1 then
        return ____exports.registerUndoAction(actions[1])
    end
    local playerIndex = actions[1].playerIndex
    assert(
        __TS__ArrayEvery(
            actions,
            function(____, a) return a.playerIndex == playerIndex end
        ),
        "All actions must be for the same player"
    )
    ____exports.registerUndoAction({handlerName = "_undoGroup", playerIndex = playerIndex, data = actions})
end
FutureUndoTranslation = "bp100:future-undo-fake-translation"
Events.on_string_translated(function(e)
    if e.localised_string ~= FutureUndoTranslation then
        return
    end
    local entry = global.futureUndoData[e.id]
    if not entry then
        return
    end
    global.futureUndoData[e.id] = nil
    local player = game.get_player(e.player_index)
    registerUndo(entry.handlerName, player, entry.data)
end)
local function doUndoEntry(player, entry)
    local handler = undoHandlers[entry.handlerName]
    if not handler then
        error("Undo handler not found: " .. entry.handlerName)
    end
    handler(player, entry.data)
end
local function doUndoEntryAtIndex(playerIndex, undoIndex)
    local playerData = global.players[playerIndex]
    local entry = playerData.undoEntries[undoIndex]
    if not entry then
        return
    end
    playerData.undoEntries[undoIndex] = nil
    playerData.nextUndoEntryIndex = undoIndex
    local player = game.get_player(playerIndex)
    assert(player)
    doUndoEntry(player, entry)
end
____exports.UndoHandler(
    "_undoGroup",
    function(player, actions)
        for i = #actions, 1, -1 do
            doUndoEntry(player, actions[i])
        end
    end
)
function ____exports.onUndoReferenceBuilt(playerIndex, entity)
    local position = entity.position
    entity.destroy()
    local index = position.x
    if index ~= floor(index) or position.y ~= 0 then
        return
    end
    doUndoEntryAtIndex(playerIndex, index)
end
function ____exports._simulateUndo(player, index)
    if index == nil then
        index = _lastUndoIndex or error("No undo action to simulate")
    end
    doUndoEntryAtIndex(player.index, index)
    _lastUndoIndex = nil
end
function ____exports.performUndoAction(action)
    doUndoEntry(
        game.get_player(action.playerIndex),
        action
    )
end
return ____exports
